@@ -47,21 +47,15 @@ typedef enum {
 	EV_FILE_EXPORTER_CAN_NUMBER_UP    = 1 << 8
 } EvFileExporterCapabilities;
 
-typedef enum {
-	EV_FILE_EXPORTER_PORTRAIT,
-	EV_FILE_EXPORTER_LANDSCAPE
-} EvFileExporterOrientation;
-
 typedef struct {
-	EvFileExporterFormat      format;
-	const gchar              *filename;
-	gint                      first_page;
-	gint                      last_page;
-	gdouble                   paper_width;
-	gdouble                   paper_height;
-	EvFileExporterOrientation orientation;
-	gboolean                  duplex;
-	gint                      pages_per_sheet;
+	EvFileExporterFormat format;
+	const gchar         *filename;
+	gint                 first_page;
+	gint                 last_page;
+	gdouble              paper_width;
+	gdouble              paper_height;
+	gboolean             duplex;
+	gint                 pages_per_sheet;
 } EvFileExporterContext;
 
 #define EV_TYPE_FILE_EXPORTER            (ev_file_exporter_get_type ())
@@ -80,8 +74,10 @@ struct _EvFileExporterIface {
         /* Methods  */
         void                       (* begin)            (EvFileExporter        *exporter,
 							 EvFileExporterContext *fc);
+	void                       (* begin_page)       (EvFileExporter        *exporter);
         void                       (* do_page)          (EvFileExporter        *exporter,
 							 EvRenderContext       *rc);
+	void                       (* end_page)         (EvFileExporter        *exporter);
         void                       (* end)              (EvFileExporter        *exporter);
 	EvFileExporterCapabilities (* get_capabilities) (EvFileExporter        *exporter);
 };
@@ -89,8 +85,10 @@ struct _EvFileExporterIface {
 GType                      ev_file_exporter_get_type         (void) G_GNUC_CONST;
 void                       ev_file_exporter_begin            (EvFileExporter        *exporter,
 							      EvFileExporterContext *fc);
+void                       ev_file_exporter_begin_page       (EvFileExporter        *exporter);
 void                       ev_file_exporter_do_page          (EvFileExporter        *exporter,
 							      EvRenderContext       *rc);
+void                       ev_file_exporter_end_page         (EvFileExporter        *exporter);
 void                       ev_file_exporter_end              (EvFileExporter        *exporter);
 EvFileExporterCapabilities ev_file_exporter_get_capabilities (EvFileExporter        *exporter);
 
