@@ -3,16 +3,17 @@
 #include "ev-document-thumbnails.h"
 #include "ev-document-links.h"
 #include "ev-document-images.h"
+#include "ev-document-forms.h"
+#include "ev-file-exporter.h"
 #include "ev-document-factory.h"
 #include "ev-document-misc.h"
 #include "ev-file-helpers.h"
 #include "ev-document-fonts.h"
-#include "ev-selection.h"
 #include "ev-async-renderer.h"
-#include "ev-file-exporter.h"
-#include "ev-window.h"
 
+#include <errno.h>
 #include <glib/gstdio.h>
+#include <glib/gi18n.h>
 #include <unistd.h>
 #include <libgnomevfs/gnome-vfs-uri.h>
 #include <libgnomevfs/gnome-vfs-utils.h>
@@ -28,16 +29,23 @@ static void ev_job_thumbnail_init       (EvJobThumbnail      *job);
 static void ev_job_thumbnail_class_init (EvJobThumbnailClass *class);
 static void ev_job_load_init    	(EvJobLoad	     *job);
 static void ev_job_load_class_init 	(EvJobLoadClass	     *class);
+static void ev_job_save_init            (EvJobSave           *job);
+static void ev_job_save_class_init      (EvJobSaveClass      *class);
 static void ev_job_print_init           (EvJobPrint          *job);
 static void ev_job_print_class_init     (EvJobPrintClass     *class);
 
-enum
-{
+enum {
 	FINISHED,
 	LAST_SIGNAL
 };
 
+enum {
+	PAGE_READY,
+	RENDER_LAST_SIGNAL
+};
+
 static guint job_signals[LAST_SIGNAL] = { 0 };
+static guint job_render_signals[RENDER_LAST_SIGNAL] = { 0 };
 
 G_DEFINE_TYPE (EvJob, ev_job, G_TYPE_OBJECT)
 G_DEFINE_TYPE (EvJobLinks, ev_job_links, EV_TYPE_JOB)
@@ -45,6 +53,7 @@ G_DEFINE_TYPE (EvJobRender, ev_job_render, EV_TYPE_JOB)
 G_DEFINE_TYPE (EvJobThumbnail, ev_job_thumbnail, EV_TYPE_JOB)
 G_DEFINE_TYPE (EvJobFonts, ev_job_fonts, EV_TYPE_JOB)
 G_DEFINE_TYPE (EvJobLoad, ev_job_load, EV_TYPE_JOB)
+G_DEFINE_TYPE (EvJobSave, ev_job_save, EV_TYPE_JOB)
 G_DEFINE_TYPE (EvJobPrint, ev_job_print, EV_TYPE_JOB)
 
 static void ev_job_init (EvJob *job) { /* Do Nothing */ }
@@ -151,6 +160,15 @@ ev_job_render_class_init (EvJobRenderClass *class)
 
 	oclass = G_OBJECT_CLASS (class);
 
+	job_render_signals [PAGE_READY] =
+		g_signal_new ("page-ready",
+			      EV_TYPE_JOB_RENDER,
+			      G_SIGNAL_RUN_LAST,
+			      G_STRUCT_OFFSET (EvJobRenderClass, page_ready),
+			      NULL, NULL,
+			      g_cclosure_marshal_VOID__VOID,
+			      G_TYPE_NONE, 0);
+
 	oclass->dispose = ev_job_render_dispose;
 }
 
@@ -263,8 +281,10 @@ ev_job_render_new (EvDocument      *document,
 		   gint             width,
 		   gint             height,
 		   EvRectangle     *selection_points,
+		   EvSelectionStyle selection_style,
 		   GdkColor        *text,
 		   GdkColor        *base,
+		   gboolean 	    include_forms,
 		   gboolean         include_links,
 		   gboolean         include_images,
 		   gboolean         include_text,
@@ -282,8 +302,10 @@ ev_job_render_new (EvDocument      *document,
 	job->rc = g_object_ref (rc);
 	job->target_width = width;
 	job->target_height = height;
+	job->selection_style = selection_style;
 	job->text = *text;
 	job->base = *base;
+	job->include_forms = include_forms;
 	job->include_links = include_links;
 	job->include_images = include_images;
 	job->include_text = include_text;
@@ -309,10 +331,30 @@ render_finished_cb (EvDocument      *document,
 
 	/* FIXME: ps backend should be ported to cairo */
 	job->surface = ev_document_misc_surface_from_pixbuf (pixbuf);
+	job->page_ready = TRUE;
+	g_signal_emit (job, job_render_signals[PAGE_READY], 0);
 	EV_JOB (job)->finished = TRUE;
 	ev_job_finished (EV_JOB (job));
 }
 
+static gboolean
+notify_page_ready (EvJobRender *job)
+{
+	g_signal_emit (job, job_render_signals[PAGE_READY], 0);
+
+	return FALSE;
+}
+
+static void
+ev_job_render_page_ready (EvJobRender *job)
+{
+	job->page_ready = TRUE;
+	g_idle_add_full (G_PRIORITY_HIGH_IDLE,
+			 (GSourceFunc)notify_page_ready,
+			 g_object_ref (job),
+			 (GDestroyNotify)g_object_unref);
+}
+
 void
 ev_job_render_run (EvJobRender *job)
 {
@@ -330,32 +372,41 @@ ev_job_render_run (EvJobRender *job)
 		ev_document_fc_mutex_lock ();
 		
 		job->surface = ev_document_render (EV_JOB (job)->document, job->rc);
-		if (job->include_links && EV_IS_DOCUMENT_LINKS (EV_JOB (job)->document))
-			job->link_mapping =
-				ev_document_links_get_links (EV_DOCUMENT_LINKS (EV_JOB (job)->document),
-							     job->rc->page);
-		if (job->include_images && EV_IS_DOCUMENT_IMAGES (EV_JOB (job)->document))
-			job->image_mapping =
-				ev_document_images_get_images (EV_DOCUMENT_IMAGES (EV_JOB (job)->document),
-							       job->rc->page);
-		if (job->include_text && EV_IS_SELECTION (EV_JOB (job)->document))
-			job->text_mapping =
-				ev_selection_get_selection_map (EV_SELECTION (EV_JOB (job)->document),
-								job->rc);
 		if (job->include_selection && EV_IS_SELECTION (EV_JOB (job)->document)) {
 			ev_selection_render_selection (EV_SELECTION (EV_JOB (job)->document),
 						       job->rc,
 						       &(job->selection),
 						       &(job->selection_points),
 						       NULL,
+						       job->selection_style,
 						       &(job->text), &(job->base));
 			job->selection_region =
 				ev_selection_get_selection_region (EV_SELECTION (EV_JOB (job)->document),
 								   job->rc,
+								   job->selection_style,
 								   &(job->selection_points));
 		}
+
+		ev_job_render_page_ready (job);
 		
 		ev_document_fc_mutex_unlock ();
+		
+		if (job->include_text && EV_IS_SELECTION (EV_JOB (job)->document))
+			job->text_mapping =
+				ev_selection_get_selection_map (EV_SELECTION (EV_JOB (job)->document),
+								job->rc);
+		if (job->include_links && EV_IS_DOCUMENT_LINKS (EV_JOB (job)->document))
+			job->link_mapping =
+				ev_document_links_get_links (EV_DOCUMENT_LINKS (EV_JOB (job)->document),
+							     job->rc->page);
+		if (job->include_forms && EV_IS_DOCUMENT_FORMS (EV_JOB (job)->document))
+			job->form_field_mapping =
+				ev_document_forms_get_form_fields (EV_DOCUMENT_FORMS (EV_JOB(job)->document),
+								   job->rc->page);
+		if (job->include_images && EV_IS_DOCUMENT_IMAGES (EV_JOB (job)->document))
+			job->image_mapping =
+				ev_document_images_get_images (EV_DOCUMENT_IMAGES (EV_JOB (job)->document),
+							       job->rc->page);
 		EV_JOB (job)->finished = TRUE;
 	}
 
@@ -515,6 +566,143 @@ ev_job_load_run (EvJobLoad *job)
 	EV_JOB (job)->finished = TRUE;
 }
 
+static void ev_job_save_init (EvJobSave *job) { /* Do Nothing */ }
+
+static void
+ev_job_save_dispose (GObject *object)
+{
+	EvJobSave *job = EV_JOB_SAVE (object);
+
+	if (job->uri) {
+		g_free (job->uri);
+		job->uri = NULL;
+	}
+
+	if (job->document_uri) {
+		g_free (job->document_uri);
+		job->document_uri = NULL;
+	}
+
+	if (job->error) {
+		g_error_free (job->error);
+		job->error = NULL;
+	}
+
+	(* G_OBJECT_CLASS (ev_job_save_parent_class)->dispose) (object);
+}
+
+static void
+ev_job_save_class_init (EvJobSaveClass *class)
+{
+	GObjectClass *oclass;
+
+	oclass = G_OBJECT_CLASS (class);
+
+	oclass->dispose = ev_job_save_dispose;
+}
+
+EvJob *
+ev_job_save_new (EvDocument  *document,
+		 const gchar *uri,
+		 const gchar *document_uri)
+{
+	EvJobSave *job;
+
+	job = g_object_new (EV_TYPE_JOB_SAVE, NULL);
+
+	EV_JOB (job)->document = g_object_ref (document);
+	job->uri = g_strdup (uri);
+	job->document_uri = g_strdup (document_uri);
+	job->error = NULL;
+
+	return EV_JOB (job);
+}
+
+void
+ev_job_save_run (EvJobSave *job)
+{
+	gint   fd;
+	gchar *filename;
+	gchar *tmp_filename;
+	gchar *local_uri;
+	
+	filename = ev_tmp_filename ("saveacopy");
+	tmp_filename = g_strdup_printf ("%s.XXXXXX", filename);
+	g_free (filename);
+
+	fd = g_mkstemp (tmp_filename);
+	if (fd == -1) {
+		gchar *display_name;
+		gint   save_errno = errno;
+
+		display_name = g_filename_display_name (tmp_filename);
+		g_set_error (&(job->error),
+			     G_FILE_ERROR,
+			     g_file_error_from_errno (save_errno),
+			     _("Failed to create file “%s”: %s"),
+			     display_name, g_strerror (save_errno));
+		g_free (display_name);
+		g_free (tmp_filename);
+
+		return;
+	}
+
+	ev_document_doc_mutex_lock ();
+
+	/* Save document to temp filename */
+	local_uri = g_filename_to_uri (tmp_filename, NULL, NULL);
+	ev_document_save (EV_JOB (job)->document, local_uri, &(job->error));
+	close (fd);
+
+	ev_document_doc_mutex_unlock ();
+
+	if (job->error) {
+		g_free (local_uri);
+		return;
+	}
+
+	/* If original document was compressed,
+	 * compress it again before saving
+	 */
+	if (g_object_get_data (G_OBJECT (EV_JOB (job)->document),
+			       "uri-uncompressed")) {
+		EvCompressionType ctype = EV_COMPRESSION_NONE;
+		const gchar      *ext;
+		gchar            *uri_comp;
+		
+		ext = g_strrstr (job->document_uri, ".gz");
+		if (ext && g_ascii_strcasecmp (ext, ".gz") == 0)
+			ctype = EV_COMPRESSION_GZIP;
+		
+		ext = g_strrstr (job->document_uri, ".bz2");
+		if (ext && g_ascii_strcasecmp (ext, ".bz2") == 0)
+			ctype = EV_COMPRESSION_BZIP2;
+
+		uri_comp = ev_file_compress (local_uri, ctype, &(job->error));
+		g_free (local_uri);
+		ev_tmp_filename_unlink (tmp_filename);
+
+		if (!uri_comp || job->error) {
+			local_uri = NULL;
+		} else {
+			local_uri = uri_comp;
+		}
+	}
+
+	g_free (tmp_filename);
+	
+	if (job->error) {
+		g_free (local_uri);
+		return;
+	}
+
+	if (!local_uri)
+		return;
+
+	ev_xfer_uri_simple (local_uri, job->uri, &(job->error));
+	ev_tmp_uri_unlink (local_uri);
+}
+
 EvJob *
 ev_job_print_new (EvDocument    *document,
 		  const gchar   *format,
@@ -523,6 +711,7 @@ ev_job_print_new (EvDocument    *document,
 		  EvPrintRange  *ranges,
 		  gint           n_ranges,
 		  EvPrintPageSet page_set,
+		  gint           pages_per_sheet,
 		  gint           copies,
 		  gdouble        collate,
 		  gdouble        reverse)
@@ -545,6 +734,8 @@ ev_job_print_new (EvDocument    *document,
 	job->n_ranges = n_ranges;
 
 	job->page_set = page_set;
+
+	job->pages_per_sheet = pages_per_sheet;
 	
 	job->copies = copies;
 	job->collate = collate;
@@ -590,21 +781,6 @@ ev_print_job_get_last_page (EvJobPrint *job)
 	return MIN (max_page, last_page);
 }
 
-static gboolean
-ev_print_job_print_page_in_range (EvJobPrint *job,
-				  gint        page)
-{
-	gint i;
-
-	for (i = 0; i < job->n_ranges; i++) {
-		if (page >= job->ranges[i].start &&
-		    page <= job->ranges[i].end)
-			return TRUE;
-	}
-
-	return FALSE;
-}
-
 static gboolean
 ev_print_job_print_page_in_set (EvJobPrint *job,
 				gint        page)
@@ -621,26 +797,74 @@ ev_print_job_print_page_in_set (EvJobPrint *job,
 	return FALSE;
 }
 
-static void
-ev_job_print_do_page (EvJobPrint *job, gint page)
+static gint *
+ev_job_print_get_page_list (EvJobPrint *job,
+			    gint       *n_pages)
 {
-	EvDocument      *document = EV_JOB (job)->document;
-	EvRenderContext *rc;
+	gint  i, j, page, max_page;
+	gint  pages = 0;
+	gint *page_list;
 
-	rc = ev_render_context_new (0, page, 1.0);
-	ev_file_exporter_do_page (EV_FILE_EXPORTER (document), rc);
-	g_object_unref (rc);
+	max_page = ev_document_get_n_pages (EV_JOB (job)->document) - 1;
+
+	for (i = 0; i < job->n_ranges; i++) {
+		gint rsize;
+		gint start, end;
+
+		if (job->ranges[i].start > max_page)
+			continue;
+		
+		start = job->ranges[i].start + 1;
+		end = job->ranges[i].end <= max_page ? job->ranges[i].end + 1 : max_page + 1;
+		rsize = end - start + 1;
+
+		switch (job->page_set) {
+		        case EV_PRINT_PAGE_SET_EVEN:
+				pages += start % 2 == 0 ? (rsize / 2) + (rsize % 2) : (rsize / 2);
+				break;
+		        case EV_PRINT_PAGE_SET_ODD:
+				pages += start % 2 != 0 ? (rsize / 2) + (rsize % 2) : (rsize / 2);
+				break;
+		        default:
+				pages += rsize;
+				break;
+		}
+	}
+
+	*n_pages = pages;
+
+	if (pages == 0)
+		return NULL;
+
+	page_list = g_new (gint, pages);
+
+	page = 0;
+	for (i = 0; i < job->n_ranges; i++) {
+		for (j = job->ranges[i].start; j <= job->ranges[i].end; j++) {
+			if (j > max_page)
+				break;
+		
+			if (ev_print_job_print_page_in_set (job, j + 1))
+				page_list[page++] = j;
+		}
+	}
+
+	return page_list;
 }
 
 void
 ev_job_print_run (EvJobPrint *job)
 {
-	EvDocument *document = EV_JOB (job)->document;
-	gint        fd;
-	gint        last_page;
-	gint        first_page;
-	gint        i;
-	gchar      *filename;
+	EvDocument            *document = EV_JOB (job)->document;
+	EvFileExporterContext  fc;
+	EvRenderContext       *rc;
+	gint                   fd;
+	gint                  *page_list;
+	gint                   n_pages;
+	gint                   last_page;
+	gint                   first_page;
+	gint                   i, j;
+	gchar                 *filename;
 	
 	g_return_if_fail (EV_IS_JOB_PRINT (job));
 
@@ -660,49 +884,58 @@ ev_job_print_run (EvJobPrint *job)
 		return;
 	}
 
+	page_list = ev_job_print_get_page_list (job, &n_pages);
+	if (n_pages == 0) {
+		close (fd);
+		EV_JOB (job)->finished = TRUE;
+		return;
+	}
+
 	first_page = ev_print_job_get_first_page (job);
 	last_page = ev_print_job_get_last_page (job);
 
+	fc.format = g_ascii_strcasecmp (job->format, "pdf") == 0 ?
+		EV_FILE_FORMAT_PDF : EV_FILE_FORMAT_PS;
+	fc.filename = job->temp_file;
+	fc.first_page = MIN (first_page, last_page);
+	fc.last_page = MAX (first_page, last_page);
+	fc.paper_width = job->width;
+	fc.paper_height = job->height;
+	fc.duplex = FALSE;
+	fc.pages_per_sheet = MAX (1, job->pages_per_sheet);
+
+	rc = ev_render_context_new (0, 0, 1.0);
+
 	ev_document_doc_mutex_lock ();
-	ev_file_exporter_begin (EV_FILE_EXPORTER (document),
-				g_ascii_strcasecmp (job->format, "pdf") == 0 ?
-				EV_FILE_FORMAT_PDF : EV_FILE_FORMAT_PS,
-				job->temp_file,
-				MIN (first_page, last_page),
-				MAX (first_page, last_page),
-				job->width, job->height, FALSE);
-	ev_document_doc_mutex_unlock ();
+	ev_file_exporter_begin (EV_FILE_EXPORTER (document), &fc);
 
 	for (i = 0; i < job->copies; i++) {
 		gint page, step;
+		gint n_copies;
 		
-		step = job->reverse ? -1 : 1;
-		page = job->reverse ? last_page : first_page;
-		
-		while ((job->reverse && (page >= first_page)) ||
-		       (!job->reverse && (page <= last_page))) {
-			gint n_pages = 1;
-			gint j;
+		step = job->reverse ? -1 * job->pages_per_sheet : job->pages_per_sheet;
+		page = job->reverse ? (n_pages / job->pages_per_sheet) * job->pages_per_sheet  : 0;
+		n_copies = job->collate ? job->copies : 1;
 
-			if (job->n_ranges > 0 &&
-			    !ev_print_job_print_page_in_range (job, page)) {
-				page += step;
-				continue;
+		while ((job->reverse && (page >= 0)) || (!job->reverse && (page < n_pages))) {
+			gint k;
+
+			ev_file_exporter_begin_page (EV_FILE_EXPORTER (document));
+			
+			for (j = 0; j < job->pages_per_sheet; j++) {
+				gint p = page + j;
+				
+				if (p < 0 || p >= n_pages)
+					break;
+
+				ev_render_context_set_page (rc, page_list[p]);
+
+				for (k = 0; k < n_copies; k++) {
+					ev_file_exporter_do_page (EV_FILE_EXPORTER (document), rc);
+				}
 			}
 
-			if (!ev_print_job_print_page_in_set (job, page + 1)) {
-				page += step;
-				continue;
-			}
-
-			if (job->collate)
-				n_pages = job->copies;
-
-			for (j = 0; j < n_pages; j++) {
-				ev_document_doc_mutex_lock ();
-				ev_job_print_do_page (job, page);
-				ev_document_doc_mutex_unlock ();
-			}
+			ev_file_exporter_end_page (EV_FILE_EXPORTER (document));
 
 			page += step;
 		}
@@ -711,11 +944,12 @@ ev_job_print_run (EvJobPrint *job)
 			break;
 	}
 
-	ev_document_doc_mutex_lock ();
 	ev_file_exporter_end (EV_FILE_EXPORTER (document));
 	ev_document_doc_mutex_unlock ();
-
+	
+	g_free (page_list);
 	close (fd);
+	g_object_unref (rc);
 	
 	EV_JOB (job)->finished = TRUE;
 }
